# -*- coding: utf-8 -*-
__all__ = ('FormatDatabase',)

from types import MappingProxyType
from typing import Mapping, Dict, Set, Any

import yaml

from .msg import MsgFormat
from .srv import SrvFormat
from .action import ActionFormat
from .package import PackageDatabase


class FormatDatabase:
    """
    An immutable database of ROS definitions that maintains the parsed
    contents of :code:`.msg`, :code:`.srv`, and :code:`.action` files
    for messages, services, and action definitions.
    Note that implicit message definitions, i.e., those associated with a
    service or action definition, are also represented by the database.

    Attributes
    ----------
    messages: Mapping[str, MsgFormat]
        An immutable mapping from message name to its definition.
    services: Mapping[str, SrvFormat]
        An immutable mapping from service name to definition.
    actions: Mapping[str, ActionFormat]
        An immutable mapping from action name to definition.
    """
    @staticmethod
    def build(db: PackageDatabase) -> 'FormatDatabase':
        """Constructs a format database from a given package database."""
        messages: Set[MsgFormat] = set()
        services: Set[SrvFormat] = set()
        actions: Set[ActionFormat] = set()

        for package in db.values():
            messages.update(package.messages)
            services.update(package.services)
            actions.update(package.actions)

            for service in package.services:
                if service.request:
                    messages.add(service.request)
                if service.response:
                    messages.add(service.response)

            for action in package.actions:
                if action.goal:
                    messages.add(action.goal)
                if action.result:
                    messages.add(action.result)
                if action.feedback:
                    messages.add(action.feedback)

        return FormatDatabase(messages, services, actions)

    def __init__(self,
                 messages: Set[MsgFormat],
                 services: Set[SrvFormat],
                 actions: Set[ActionFormat]
                 ) -> None:
        self.__messages: Mapping[str, MsgFormat] = MappingProxyType({
            f.fullname: f for f in messages})
        self.__services: Mapping[str, SrvFormat] = MappingProxyType({
            f.fullname: f for f in services})
        self.__actions: Mapping[str, ActionFormat] = MappingProxyType({
            f.fullname: f for f in actions})

    @property
    def messages(self) -> Mapping[str, MsgFormat]:
        return self.__messages

    @property
    def services(self) -> Mapping[str, SrvFormat]:
        return self.__services

    @property
    def actions(self) -> Mapping[str, ActionFormat]:
        return self.__actions

    def to_dict(self) -> Dict[str, Any]:
        """Returns a JSON description of this database."""
        return {'messages': [m.to_dict() for m in self.__messages.values()],
                'services': [s.to_dict() for s in self.__services.values()],
                'actions': [a.to_dict() for a in self.__actions.values()]}

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> 'FormatDatabase':
        """Loads a format database from a JSON document."""
        msg = {MsgFormat.from_dict(dd) for dd in d['messages']}
        srv = {SrvFormat.from_dict(dd) for dd in d['services']}
        action = {ActionFormat.from_dict(dd) for dd in d['actions']}
        return FormatDatabase(msg, srv, action)

    def save(self, fn: str) -> None:
        """Saves the contents of this format database to disk."""
        with open(fn, 'w') as f:
            yaml.dump(self.to_dict(), f, default_flow_style=False)

    @classmethod
    def load(cls, fn: str) -> 'FormatDatabase':
        """Loads a format database from a given file on disk."""
        with open(fn, 'r') as f:
            return cls.from_dict(yaml.safe_load(f))



from typing import Iterator, Tuple
import os
import contextlib
import logging
import time

import pytest

import roswire
import roswire.exceptions
from dockerblade import Shell, FileSystem
from roswire import ROSWire, System, SystemDescription
from roswire.proxy import ROSCore, Container
from roswire.description import SystemDescription
from roswire.definitions import TypeDatabase, FormatDatabase, PackageDatabase

DIR_TEST = os.path.dirname(__file__)


def skip_if_on_travis(f):
    if os.environ.get('TRAVIS') == 'true':
        return pytest.mark.skipif(f, reason='skipping test on Travis')
    else:
        return f


def load_hello_world_type_db() -> TypeDatabase:
    fn_db_format = os.path.join(DIR_TEST,
                                'format-databases/helloworld.formats.yml')
    db_format = FormatDatabase.load(fn_db_format)
    return TypeDatabase.build(db_format)


def load_hello_world_description() -> SystemDescription:
    fn_db_format = os.path.join(DIR_TEST,
                                'format-databases/helloworld.formats.yml')
    db_format = FormatDatabase.load(fn_db_format)
    db_type = TypeDatabase.build(db_format)
    desc = SystemDescription(sha256='foo',
                             types=db_type,
                             formats=db_format,
                             packages=PackageDatabase([]))
    return desc


@contextlib.contextmanager
def build_ardu() -> Iterator[Tuple[System, ROSCore]]:
    rsw = ROSWire()
    with rsw.launch('brass') as sut:
        with sut.roscore() as ros:
            time.sleep(5)
            yield (sut, ros)


@contextlib.contextmanager
def build_hello_world() -> Iterator[Tuple[System, ROSCore]]:
    rsw = ROSWire()
    image = 'roswire/helloworld:buggy'
    desc = load_hello_world_description()
    with rsw.launch(image, desc) as sut:
        with sut.roscore() as ros:
            time.sleep(5)
            yield (sut, ros)


@contextlib.contextmanager
def build_shell_proxy() -> Iterator[Shell]:
    rsw = ROSWire()
    image = 'brass'
    desc = SystemDescription(image, [], [], [])
    with rsw.launch(image, desc) as sut:
        yield sut.shell


@contextlib.contextmanager
def build_file_proxy() -> Iterator[FileSystem]:
    rsw = ROSWire()
    image = 'brass'
    desc = SystemDescription(image, [], [], [])
    with rsw.launch(image, desc) as sut:
        yield sut.files


@contextlib.contextmanager
def build_file_and_shell_proxy() -> Iterator[Tuple[FileSystem, Shell]]:
    rsw = ROSWire()
    image = 'brass'
    desc = SystemDescription(image, [], [], [])
    with rsw.launch(image, desc) as sut:
        yield sut.files, sut.shell


@skip_if_on_travis
@pytest.mark.parametrize('sut', ['fetch'], indirect=True)
def test_parameters(sut):
    with sut.roscore() as ros:
        assert ros.topic_to_type == {'/rosout': 'rosgraph_msgs/Log',
                                     '/rosout_agg': 'rosgraph_msgs/Log'}

        assert '/rosversion' in ros.parameters
        assert '/rosdistro' in ros.parameters

        assert '/hello' not in ros.parameters
        ros.parameters['/hello'] = 'world'
        assert '/hello' in ros.parameters
        assert ros.parameters['/hello'] == 'world'

        del ros.parameters['/hello']
        assert 'hello' not in ros.parameters
        with pytest.raises(KeyError):
            ros.parameters['/hello']



# -*- coding: utf-8 -*-
__all__ = ('OpCode', 'Compression', 'BagMessage', 'ChunkConnection', 'Chunk',
           'ConnectionInfo', 'BagHeader', 'IndexEntry', 'Index')

from typing import Dict, Optional, Tuple, List
from enum import Enum

import attr

from ..definitions.base import Time
from ..definitions.msg import Message
from ..util import tuple_from_iterable


class OpCode(Enum):
    MESSAGE_DATA = b'\x02'
    HEADER = b'\x03'
    INDEX_DATA = b'\x04'
    CHUNK = b'\x05'
    CHUNK_INFO = b'\x06'
    CONNECTION_INFO = b'\x07'

    @property
    def hex(self) -> str:
        return f'0x{self.value.hex()}'


class Compression(Enum):
    NONE = 'none'
    BZ2 = 'bz2'


@attr.s(frozen=True, slots=True, auto_attribs=True)
class BagMessage:
    topic: str
    time: Time
    message: Message


@attr.s(frozen=True, slots=True, auto_attribs=True)
class ChunkConnection:
    """
    Attributes
    ----------
    uid: str
        The connection identifier
    count: int
        The number of messages that arrived on this connection in the chunk
    """
    uid: int
    count: int


@attr.s(frozen=True, slots=True, auto_attribs=True)
class Chunk:
    pos_record: int
    pos_data: int
    time_start: Time
    time_end: Time
    connections: Tuple[ChunkConnection, ...] = \
        attr.ib(converter=tuple_from_iterable)
    compression: Compression
    size_uncompressed: int
    size_compressed: int


@attr.s(frozen=True, slots=True, auto_attribs=True)
class ConnectionInfo:
    conn: int
    topic: str
    topic_original: str
    typ: str
    md5sum: str
    message_definition: str
    callerid: Optional[str]
    latching: Optional[str]


@attr.s(frozen=True, slots=True, auto_attribs=True)
class BagHeader:
    index_pos: int
    conn_count: int
    chunk_count: int


@attr.s(frozen=True, slots=True, auto_attribs=True)
class IndexEntry:
    time: Time
    pos: int
    offset: int


Index = Dict[int, List[IndexEntry]]



# -*- coding: utf-8 -*-
"""
This module is used to build descriptions of containerised ROS applications,
and provides a manager for persisting those descriptions to disk.
"""
__all__ = ('SystemDescription', 'SystemDescriptionManager')

from typing import Any, Dict, Sequence, Union
import base64
import os
import typing

from docker.models.images import Image as DockerImage
from loguru import logger
import attr
import yaml

from .definitions import TypeDatabase, FormatDatabase, PackageDatabase

if typing.TYPE_CHECKING:
    from .proxy import ContainerManager


@attr.s(slots=True, frozen=True, auto_attribs=True)
class SystemDescription:
    """
    An immutable description of all of the packages, types, and
    specifications (i.e., :code:`.msg`, :code:`.srv`, and :code:`.action`
    files) within a containerised ROS application.

    Attributes
    ----------
    sha256: str
        The ID of the Docker image for the application.
    sources: Sequence[str]
        The sequence of setup files that should be used to load the ROS
        workspace for the application.
    types: TypeDatabase
        A database of types for the application.
    formats: FormatDatabase
        A database of message, service and action specifications.
    packages: PackageDatabase
        A database of the packages contained within the application.
    """
    sha256: str
    sources: Sequence[str]
    types: TypeDatabase
    formats: FormatDatabase
    packages: PackageDatabase

    @staticmethod
    def from_file(fn: str) -> 'SystemDescription':
        with open(fn, 'r') as f:
            d = yaml.safe_load(f)
        return SystemDescription.from_dict(d)

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> 'SystemDescription':
        sha256: str = d['sha256']
        sources: Sequence[str] = tuple(d['sources'])
        packages = PackageDatabase.from_dict(d['packages'])
        formats = FormatDatabase.build(packages)
        types = TypeDatabase.build(formats)
        return SystemDescription(sha256=sha256,
                                 sources=sources,
                                 packages=packages,
                                 formats=formats,
                                 types=types)

    def to_dict(self) -> Dict[str, Any]:
        return {'sha256': self.sha256,
                'sources': list(self.sources),
                'packages': self.packages.to_dict()}


class SystemDescriptionManager:
    """Builds and stores descriptions of containerised ROS applications.

    Attributes
    ----------
    dir: str
        The absolute path of the directory where descriptions are stored.
    """
    def __init__(self,
                 containers: 'ContainerManager',
                 dir_cache: str
                 ) -> None:
        self.__containers = containers

        # ensure that the cache directory exists
        self.__dir_cache = dir_cache
        os.makedirs(dir_cache, exist_ok=True)

    @property
    def dir(self) -> str:
        return self.__dir_cache

    def _path_for_image_with_sources(self,
                                     image_or_tag: Union[str, DockerImage],
                                     sources: Sequence[str]
                                     ) -> str:
        sha256 = self.__containers.image_sha256(image_or_tag)
        return self._path_for_sha256_with_sources(sha256, sources)

    def _path_for_sha256_with_sources(self,
                                      sha256: str,
                                      sources: Sequence[str]
                                      ) -> str:
        unencoded_path = ':'.join([sha256] + list(sources))
        encoded_path = \
            base64.b64encode(unencoded_path.encode('utf-8')).decode('utf-8')
        return os.path.join(self.__dir_cache, encoded_path)

    def load(self,
             image_or_tag: Union[str, DockerImage],
             sources: Sequence[str]
             ) -> SystemDescription:
        """Attempts to load the description for a given Docker image from disk.

        Parameters
        ----------
        image_or_tag: Union[str, DockerImage]
            the name or object for the Docker image.
        sources: Sequence[str]
            The sequence of setup files that should be used to load the ROS
            workspace.

        Returns
        -------
        SystemDescription
            A description of the application contained within the given image.

        Raises
        ------
        FileNotFoundError
            if no description for the given image is stored on disk.
        """
        filename = self._path_for_image_with_sources(image_or_tag, sources)
        try:
            return SystemDescription.from_file(filename)
        except FileNotFoundError:
            logger.exception('failed to load description for image '
                             f'[{image_or_tag}]: file [{filename}] not found.')
            raise

    def saved(self,
              image_or_tag: Union[str, DockerImage],
              sources: Sequence[str]
              ) -> bool:
        """Determines whether an application description has been saved."""
        filename = self._path_for_image_with_sources(image_or_tag, sources)
        return os.path.exists(filename)

    def save(self, description: SystemDescription) -> None:
        """Saves a given description to disk."""
        filename = self._path_for_sha256_with_sources(description.sha256,
                                                      description.sources)
        yml = description.to_dict()
        with open(filename, 'w') as f:
            yaml.dump(yml, f, default_flow_style=False)

    def build(self,
              image_or_tag: Union[str, DockerImage],
              sources: Sequence[str],
              save: bool = True
              ) -> SystemDescription:
        """
        Builds a description of the ROS application contained within a given
        image and optionally saves that description to disk.

        Parameters
        ----------
        image_or_tag: Union[str, DockerImage]
            The name or object for the Docker image.
        sources: Sequence[str]
            The sequence of setup files that should be used to load the ROS
            workspace.
        save: bool
            If :code:`True`, the description will be saved to disk.

        Returns
        -------
        SystemDescription
            A description of the application contained within the given image.
        """
        with self.__containers.launch(image_or_tag, sources=sources) as container:  # noqa
            paths = PackageDatabase.paths(container.shell, container.files)
            db_package = PackageDatabase.from_paths(container.files, paths)
        db_format = FormatDatabase.build(db_package)
        db_type = TypeDatabase.build(db_format)

        sha256 = self.__containers.image_sha256(image_or_tag)
        description = SystemDescription(sha256=sha256,
                                        sources=sources,
                                        packages=db_package,
                                        formats=db_format,
                                        types=db_type)
        if save:
            self.save(description)
        return description

    def load_or_build(self,
                      image_or_tag: Union[str, DockerImage],
                      sources: Sequence[str],
                      save: bool = True
                      ) -> SystemDescription:
        """
        Attempts to load a description for a given image from disk, and if
        none can be found, a description will be built from scratch and
        optionally saved.

        Parameters
        ----------
        image_or_tag: Union[str, DockerImage]
            the name or object for the Docker image.
        sources: Sequence[str]
            The sequence of setup files that should be used to load the ROS
            workspace.
        save: bool
            if :code:`True`, the description will be saved to disk in the
            event that it needs to be built from scratch.

        Returns
        -------
        SystemDescription
            A description of the application contained within the given image.
        """
        if self.saved(image_or_tag, sources):
            return self.load(image_or_tag, sources)
        return self.build(image_or_tag, sources, save)



import pytest

from roswire.definitions import (TypeDatabase, FormatDatabase, MsgFormat,
                                 Time, Duration)

from test_bag import load_mavros_type_db


def test_encode_and_decode():
    db_type = load_mavros_type_db()
    Header = db_type['std_msgs/Header']
    orig = Header(seq=32, stamp=Time(secs=9781, nsecs=321), frame_id='')
    assert orig == Header.decode(orig.encode())


def test_to_and_from_dict():
    db_type = load_mavros_type_db()
    Waypoint = db_type['mavros_msgs/Waypoint']
    WaypointPushRequest = db_type['mavros_msgs/WaypointPushRequest']

    wp = Waypoint(frame=0,
                  command=16,
                  is_current=False,
                  autocontinue=False,
                  param1=0.0,
                  param2=1.0,
                  param3=2.0,
                  param4=3.0,
                  x_lat=4.0,
                  y_long=5.0,
                  z_alt=6.0)
    req = WaypointPushRequest(waypoints=[wp])

    assert req.to_dict() == {'waypoints': [
        {'frame': 0,
         'command': 16,
         'is_current': False,
         'autocontinue': False,
         'param1': 0.0,
         'param2': 1.0,
         'param3': 2.0,
         'param4': 3.0,
         'x_lat': 4.0,
         'y_long': 5.0,
         'z_alt': 6.0}
    ]}



# -*- coding: utf-8 -*-
__all__ = ('SystemState', 'SystemStateProbe')

from typing import AbstractSet, Collection, Dict, Mapping, Set, Sequence, Tuple
import xmlrpc

import attr

from .. import exceptions as exc


@attr.s(frozen=True, auto_attribs=True, slots=True)
class SystemState:
    """Provides a description of the instantaneous state of a ROS system in
    terms of its publishers, subscribers, and services.

    Attributes
    ----------
    publishers: Mapping[str, Collection[str]]
        A mapping from topics to the names of publishers to that topic.
    subscribers: Mapping[str, Collection[str]]
        A mapping from topics to the names of subscribers to that topic.
    services: Mapping[str, Collection[str]]
        A mapping from services to the names of providers of that service.
    nodes: AbstractSet[str]
        The names of all known nodes running on the system.
    topics: AbstractSet[str]
        The names of all known topics on the system with at least one
        publisher or one subscriber.
    """
    publishers: Mapping[str, Collection[str]]
    subscribers: Mapping[str, Collection[str]]
    services: Mapping[str, Collection[str]]
    nodes: AbstractSet[str] = attr.ib(init=False, repr=False)
    topics: AbstractSet[str] = attr.ib(init=False, repr=False)

    def __attrs_post_init__(self) -> None:
        nodes: Set[str] = set()
        nodes = nodes.union(*self.publishers.values())
        nodes = nodes.union(*self.subscribers.values())
        nodes = nodes.union(*self.services.values())

        topics: Set[str] = set()
        topics = topics.union(self.publishers)
        topics = topics.union(self.subscribers)

        object.__setattr__(self, 'nodes', frozenset(nodes))
        object.__setattr__(self, 'topics', frozenset(topics))


@attr.s(frozen=True, auto_attribs=True)
class SystemStateProbe:
    """Provides an interface for obtaining the instantaneous state of a ROS
    system in terms of its publishers, subscribers, and services."""
    _connection: xmlrpc.client.ServerProxy

    @classmethod
    def via_xmlrpc_connection(cls,
                              connection: xmlrpc.client.ServerProxy
                              ) -> 'SystemStateProbe':
        return SystemStateProbe(connection)

    def probe(self) -> SystemState:
        """Obtains the instantaneous state of the associated ROS system."""
        code: int
        msg: str
        result: Tuple[Sequence[Tuple[str, Sequence[str]]],
                      Sequence[Tuple[str, Sequence[str]]],
                      Sequence[Tuple[str, Sequence[str]]]]
        code, msg, result = \
            self._connection.getSystemState('roswire-probe')  # type: ignore
        if code != 1:
            raise exc.ROSWireException("probe failed!")

        publishers: Dict[str, Sequence[str]] = {}
        subscribers: Dict[str, Sequence[str]] = {}
        services: Dict[str, Sequence[str]] = {}

        for topic, publisher_names in result[0]:
            publishers[topic] = publisher_names
        for topic, subscriber_names in result[1]:
            subscribers[topic] = subscriber_names
        for topic, service_names in result[2]:
            services[topic] = service_names

        state = SystemState(publishers=publishers,
                            subscribers=subscribers,
                            services=services)
        return state

    __call__ = probe



# -*- coding: utf-8 -*-
"""
This module provides a unified interface for interacting with the catkin
and catkin_make build systems.
"""
__all__ = ('CatkinInterface', 'CatkinTools', 'CatkinMake')

from typing import Optional, List
import abc
import shlex

from loguru import logger
import attr
import dockerblade

from ..exceptions import CatkinBuildFailed, CatkinCleanFailed


class CatkinInterface(abc.ABC):
    """Provides an interface to a catkin-based workspace."""
    @property
    @abc.abstractmethod
    def directory(self) -> str:
        """The directory of this catkin workspace."""
        ...

    @abc.abstractmethod
    def clean(self,
              packages: Optional[List[str]] = None,
              orphans: bool = False,
              context: Optional[str] = None
              ) -> None:
        """Cleans all products of the catkin workspace.

        Raises
        ------
        CatkinCleanFailed:
            if the attempt to clean resulted in a non-zero return code.
        """
        ...

    @abc.abstractmethod
    def build(self,
              packages: Optional[List[str]] = None,
              no_deps: bool = False,
              pre_clean: bool = False,
              jobs: Optional[int] = None,
              cmake_args: Optional[List[str]] = None,
              make_args: Optional[List[str]] = None,
              context: Optional[str] = None,
              time_limit: Optional[int] = None
              ) -> None:
        """Attempts to build a catkin workspace.

        Raises
        ------
        CatkinBuildFailed:
            if the attempt to build resulted in a non-zero return code.
        """
        ...


@attr.s(frozen=True, slots=True, auto_attribs=True)
class CatkinTools(CatkinInterface):
    """Provides an interface to a catkin workspace created via catkin tools."""
    directory: str
    _shell: dockerblade.shell.Shell

    def clean(self,
              packages: Optional[List[str]] = None,
              orphans: bool = False,
              context: Optional[str] = None
              ) -> None:
        shell = self._shell
        command = ['catkin', 'clean', '-y']
        if orphans:
            command += ['--orphans']
        if packages:
            command += [shlex.quote(p) for p in packages]
        if not context:
            context = self.directory

        command_str = ' '.join(command)
        logger.debug(f"cleaning via: {command_str}")
        result = shell.run(command_str, cwd=context, text=True)
        duration_mins = result.duration / 60
        logger.debug("clean completed after %.2f minutes [retcode: %d]:\n%s",
                     duration_mins, result.returncode, result.output)

        if result.returncode != 0:
            assert isinstance(result.output, str)
            raise CatkinCleanFailed(result.returncode, result.output)

    def build(self,
              packages: Optional[List[str]] = None,
              no_deps: bool = False,
              pre_clean: bool = False,
              jobs: Optional[int] = None,
              cmake_args: Optional[List[str]] = None,
              make_args: Optional[List[str]] = None,
              context: Optional[str] = None,
              time_limit: Optional[int] = None
              ) -> None:
        command = ['catkin', 'build', '--no-status', '--no-notify']
        if packages:
            command += [shlex.quote(p) for p in packages]
        if no_deps:
            command += ['--no-deps']
        if pre_clean:
            command += ['--pre-clean']
        if cmake_args:
            command += cmake_args
        if make_args:
            command += make_args
        if not context:
            context = self.directory

        command_str = ' '.join(command)
        logger.debug(f"building via: {command_str}")
        result = self._shell.run(command_str,
                                 cwd=context,
                                 time_limit=time_limit,
                                 text=True)
        duration_mins = result.duration / 60
        logger.debug("build completed after %.2f minutes [retcode: %d]:\n%s",
                     duration_mins, result.returncode, result.output)

        if result.returncode != 0:
            assert isinstance(result.output, str)
            raise CatkinBuildFailed(result.returncode, result.output)


@attr.s(frozen=True, slots=True, auto_attribs=True)
class CatkinMake(CatkinInterface):
    """Provides an interface to a catkin workspace created via catkin_make."""
    directory: str
    _shell: dockerblade.shell.Shell

    def clean(self,
              packages: Optional[List[str]] = None,
              orphans: bool = False,
              context: Optional[str] = None
              ) -> None:
        raise NotImplementedError

    def build(self,
              packages: Optional[List[str]] = None,
              no_deps: bool = False,
              pre_clean: bool = False,
              jobs: Optional[int] = None,
              cmake_args: Optional[List[str]] = None,
              make_args: Optional[List[str]] = None,
              context: Optional[str] = None,
              time_limit: Optional[int] = None
              ) -> None:
        raise NotImplementedError



# -*- coding: utf-8 -*-
__all__ = (
    'canonical_name',
    'global_name',
    'namespace',
    'namespaces_of',
    'namespace_join',
    'name_is_private',
    'name_is_global',
    'name_is_legal')

from typing import List
import re

RE_NAME = re.compile(r'^[\~\/A-Za-z][\w\/]*$')


def global_name(name: str) -> str:
    """Converts a name to its global form."""
    if name_is_private(name):
        m = f'cannot convert private name [{name}] into global name'
        raise ValueError(m)
    if not name_is_global(name):
        name = '/' + name
    if name[-1] != '/':
        name = name + '/'
    return name


def canonical_name(name: str) -> str:
    """Returns the canonical form of a given name."""
    if not name:
        return ''
    if name[0] == '/':
        return '/' + '/'.join(n for n in name[1:].split('/') if n)
    return '/'.join(n for n in name.split('/') if n)


def namespaces_of(name: str) -> List[str]:
    """Returns a list of each of the namespaces for a given name."""
    if not name:
        return ['/']
    parts = [n for n in name.split('/') if n]
    return ['/'] + ['/' + '/'.join(parts[:i]) for i in range(1, len(parts))]


def name_is_legal(name: str) -> bool:
    """Determines whether a given name is a legal ROS name."""
    if name in ('', '/'):
        return True
    return RE_NAME.match(name) is not None and '//' not in name


def name_is_private(name: str) -> bool:
    """Determines whether a given name is private."""
    return name.startswith('~')


def name_is_global(name: str) -> bool:
    """Determines whether a given name is global."""
    return name.startswith('/')


def namespace(name: str) -> str:
    """Returns the corresponding namespace for a given name."""
    if name_is_private(name):
        m = f'cannot determine namespace of private name: {name}'
        raise ValueError(m)
    name = global_name(name)[:-1]
    ns, sep, name = name.rpartition('/')
    return ns if ns else '/'


def namespace_join(ns: str, name: str) -> str:
    """Concatenates a given name to a namespace."""
    if not ns or name_is_private(name) or name_is_global(name):
        return name
    if ns == '~':
        return '~' + name
    return global_name(ns) + name



# -*- coding: utf-8 -*-
__all__ = ('NodeManager', 'Node')

from typing import AbstractSet, Iterator, Mapping, Optional
from urllib.parse import urlparse
import xmlrpc.client

from loguru import logger
import dockerblade
import psutil

from .state import SystemStateProbe
from ..exceptions import ROSWireException, NodeNotFoundError


class Node:
    """Provides access to a ROS node.

    Attributes
    ----------
    api: xmlrpc.client.ServerProxy
        An XML-RPC API client for this node.
    name: str
        The fully qualified name of this node.
    url: str
        URL used to access this node from the host network.
    pid: int
        The container PID of the main process for this node.
    pid_host: int
        The host PID of the main process for this node.
    """
    def __init__(self,
                 name: str,
                 url_host_network: str,
                 shell: dockerblade.Shell
                 ) -> None:
        self.__name = name
        self.__url = url_host_network
        self.__shell = shell
        self.__pid_host: Optional[int] = None

    @property
    def api(self) -> xmlrpc.client.ServerProxy:
        return xmlrpc.client.ServerProxy(self.url)

    @property
    def name(self) -> str:
        return self.__name

    @property
    def url(self) -> str:
        return self.__url

    @property
    def pid(self) -> int:
        code, status, pid = self.api.getPid('/.roswire')  # type: ignore
        if code != 1:
            m = f"failed to obtain PID [{self.name}]: {status} (code: {code})"
            raise ROSWireException(m)
        assert isinstance(pid, int)
        assert pid > 0
        return pid

    @property
    def pid_host(self) -> int:
        if self.__pid_host is None:
            self.__pid_host = self.__shell._local_to_host_pid(self.pid)
            assert self.__pid_host is not None
        return self.__pid_host

    def is_alive(self) -> bool:
        """Determines whether this node is alive."""
        # TODO check start time to ensure this is the same process!
        try:
            return psutil.pid_exists(self.pid_host)
        except ROSWireException:
            return False

    def shutdown(self) -> None:
        """Instructs this node to shutdown."""
        self.__shell.run(f'rosnode kill {self.name}')


class NodeManager(Mapping[str, Node]):
    """Provides access to all nodes on a ROS graph."""
    def __init__(self,
                 host_ip_master: str,
                 api: xmlrpc.client.ServerProxy,
                 shell: dockerblade.Shell
                 ) -> None:
        self.__host_ip_master: str = host_ip_master
        self.__api: xmlrpc.client.ServerProxy = api
        self.__shell: dockerblade.Shell = shell
        self.__state_probe: SystemStateProbe = \
            SystemStateProbe.via_xmlrpc_connection(self.__api)

    def __get_node_names(self) -> AbstractSet[str]:
        """Fetches a list of the names of all active nodes."""
        return self.__state_probe().nodes

    def __len__(self) -> int:
        """Returns a count of the number of active nodes."""
        return len(self.__get_node_names())

    def __iter__(self) -> Iterator[str]:
        """Returns an iterator over the names of all active nodes."""
        yield from self.__get_node_names()

    def __getitem__(self, name: str) -> Node:
        """Attempts to fetch a given node.

        Parameters
        ----------
        name: str
            The name of the node.

        Returns
        -------
        Node
            An interface to the given node.

        Raises
        ------
        NodeNotFoundError
            If there is no node with the given name.
        """
        code: int
        status: str
        uri_container: str
        code, status, uri_container = \
            self.__api.lookupNode('/.roswire', name)  # type: ignore
        if code == -1:
            raise NodeNotFoundError(name)
        if code != 1:
            m = f"unexpected error when attempting to find node [{name}]: {status} (code: {code})"   # noqa: pycodestyle
            raise ROSWireException(m)

        # convert URI to host network
        port = urlparse(uri_container).port
        uri_host = f"http://{self.__host_ip_master}:{port}"
        return Node(name, uri_host, self.__shell)

    def __delitem__(self, name: str) -> None:
        """Shutdown and deregister a given node.

        Parameters
        ----------
        name: str
            The name of the node.

        Raises
        ------
        NodeNotFoundError
            no node found with given name.
        """
        try:
            node = self[name]
        except NodeNotFoundError:
            logger.exception(f"failed to delete node [{name}]: not found.")
            raise
        node.shutdown()



# -*- coding: utf-8 -*-
__all__ = ('TypeDatabase',)

from typing import (Collection, Type, Mapping, Iterator, Dict, Any, Callable,
                    BinaryIO)
from collections import OrderedDict

import attr

from .msg import MsgFormat, Field, Message
from .format import FormatDatabase
from .base import Time, get_builtin
from .decode import (is_simple,
                     read_time,
                     read_duration,
                     simple_reader,
                     string_reader,
                     complex_array_reader,
                     simple_array_reader)
from .encode import (write_time,
                     write_duration,
                     simple_writer,
                     string_writer,
                     complex_array_writer,
                     simple_array_writer)


class TypeDatabase(Mapping[str, Type[Message]]):
    """
    An immutable database of ROS types, represented as :class:`Message`
    subclasses, indexed by the fully qualified names of those types.

    Note
    ----
    Implements the set of non-destructive :class:`dict` operations as
    :class:`PackageDatabase`.
    """
    @classmethod
    def build(cls, db_format: FormatDatabase) -> 'TypeDatabase':
        formats = list(db_format.messages.values())
        formats = MsgFormat.toposort(formats)
        name_to_type: Dict[str, Type[Message]] = {}
        for fmt in formats:
            ns: Dict[str, Any] = {}
            for f in fmt.fields:
                f_base_typ: Type
                if f.base_type in name_to_type:
                    f_base_typ = name_to_type[f.base_type]
                else:
                    f_base_typ = get_builtin(f.base_type)
                converter = tuple if f.is_array else None
                ns[f.name] = attr.ib(type=f_base_typ, converter=converter)
            ns['format'] = fmt
            ns['read'] = classmethod(cls._build_read(name_to_type, fmt))
            ns['write'] = cls._build_write(name_to_type, fmt)
            md5 = fmt.md5sum(db_format.messages)
            ns['md5sum'] = classmethod(lambda cls, md5=md5: md5)
            t: Type[Message] = type(fmt.name, (Message,), ns)
            t = attr.s(t, frozen=True, slots=True)
            name_to_type[fmt.fullname] = t
        return TypeDatabase(name_to_type.values())

    @classmethod
    def _build_read(cls,
                    name_to_type: Mapping[str, Type[Message]],
                    fmt: MsgFormat
                    ) -> Callable[[Type[Message], BinaryIO], Message]:
        """Builds a reader for a given message format."""
        def get_factory(field: Field) -> Callable[[BinaryIO], Any]:
            if field.is_simple:
                return simple_reader(field.typ)
            if field.typ == 'time':
                return read_time
            if field.typ == 'duration':
                return read_duration
            if field.typ == 'string':
                return string_reader(field.length)
            if field.is_array and is_simple(field.base_type):
                return simple_array_reader(field.base_type, field.length)
            if field.is_array and not is_simple(field.base_type):
                entry_factory: Callable[[BinaryIO], Any]
                if field.base_type == 'time':
                    entry_factory = read_time
                elif field.base_type == 'duration':
                    entry_factory = read_duration
                # FIXME how about arrays of fixed-length strings?
                elif field.base_type == 'string':
                    entry_factory = string_reader()
                elif field.base_type in name_to_type:
                    entry_factory = name_to_type[field.base_type].read
                else:
                    m = "unable to find factory for base type: {}"
                    m = m.format(field.base_type)
                    raise Exception(m)
                return complex_array_reader(entry_factory, field.length)
            if field.typ in name_to_type:
                return name_to_type[field.typ].read
            m = "unable to find factory for field: {field.name} [{field.typ}]"
            raise Exception(m)

        fields: OrderedDict[str, Callable[[BinaryIO], Any]] = OrderedDict()
        for field in fmt.fields:
            fields[field.name] = get_factory(field)

        def reader(cls: Type[Message], b: BinaryIO) -> Message:
            values: Dict[str, Any] = {}
            for name, factory in fields.items():
                values[name] = factory(b)
            return cls(**values)  # type: ignore

        return reader

    @classmethod
    def _build_write(cls,
                     name_to_type: Mapping[str, Type[Message]],
                     fmt: MsgFormat
                     ) -> Callable[[Any, BinaryIO], None]:
        """Builds a write for a given message format."""
        def get_field_writer(field: Field) -> Callable[[Any, BinaryIO], None]:
            if field.is_simple:
                return simple_writer(field.typ)
            if field.typ == 'time':
                return write_time
            if field.typ == 'duration':
                return write_duration
            if field.typ == 'string':
                return string_writer(field.length)
            if field.is_array and is_simple(field.base_type):
                return simple_array_writer(field.base_type, field.length)
            if field.is_array and not is_simple(field.base_type):
                entry_writer: Callable[[Any, BinaryIO], None]
                if field.base_type == 'time':
                    entry_writer = write_time
                elif field.base_type == 'duration':
                    entry_writer = write_duration
                # FIXME how about arrays of fixed-length strings?
                elif field.base_type == 'string':
                    entry_writer = string_writer()
                elif field.base_type in name_to_type:
                    entry_writer = name_to_type[field.base_type].write
                else:
                    raise Exception(f"unable to find writer: {field.typ}")
                return complex_array_writer(entry_writer, field.length)
            if field.typ in name_to_type:
                return name_to_type[field.typ].write
            m = "unable to find writer for field: {field.name} [{field.typ}]"
            raise Exception(m)

        field_writers: OrderedDict[str, Callable[[Any, BinaryIO], None]] = \
            OrderedDict()
        for field in fmt.fields:
            field_writers[field.name] = get_field_writer(field)

        def writer(self: Message, b: BinaryIO) -> None:
            for name, field_writer in field_writers.items():
                field_writer(getattr(self, name), b)

        return writer

    def __init__(self, types: Collection[Type[Message]]) -> None:
        self.__contents: Dict[str, Type[Message]] = \
            {t.format.fullname: t for t in types}

    def __len__(self) -> int:
        return len(self.__contents)

    def __iter__(self) -> Iterator[str]:
        return self.__contents.__iter__()

    def __getitem__(self, name: str) -> Type[Message]:
        return self.__contents[name]

    def to_dict(self, message: Message) -> Dict[str, Any]:
        return message.to_dict()

    def _from_dict_value(self, field: Field, val: Any) -> Any:
        if field.is_array:
            fmt_item: MsgFormat = self[field.base_type].format
            return [self.from_dict(fmt_item, dd) for dd in val]
        if field.typ == 'time':
            return Time.from_dict(val)
        # NOTE covers simple values (e.g., str, int, float, bool)
        return val

    def from_dict(self, fmt: MsgFormat, d: Dict[str, Any]) -> Message:
        typ: Type[Message] = self[fmt.fullname]
        args: Dict[str, Any] = {f.name: self._from_dict_value(f, d[f.name])
                                for f in fmt.fields}
        return typ(**args)  # type: ignore



# -*- coding: utf-8 -*-
__all__ = ('Service', 'ServiceManager')

from typing import AbstractSet, Iterator, Mapping, Optional
from urllib.parse import urlparse
import xmlrpc.client

import attr
import dockerblade
import yaml

from .state import SystemStateProbe
from .. import exceptions
from ..definitions import Message, SrvFormat, MsgFormat
from ..description import SystemDescription


@attr.s(slots=True, auto_attribs=True)
class Service:
    """Provides access to a ROS service.

    Attributes
    ----------
    name: str
        The fully qualified name of this service.
    url: str
        The URL of this service.
    format: SrvFormat
        The :code:`.srv` definition for this service.
    """
    name: str
    url: str
    format: SrvFormat
    _description: SystemDescription
    _shell: dockerblade.Shell

    def call(self, message: Optional[Message] = None) -> Optional[Message]:
        """Calls this service.

        Parameters
        ----------
        message: Message, optional
            The message, if any, that should be sent to the service.

        Returns
        -------
        Optional[Message]
            The reply produced by the service, if any.
        """
        if not message:
            yml = '{}'
        else:
            yml = yaml.dump(message.to_dict())
        command = f"rosservice call {self.name} '{yml}'"
        try:
            output = self._shell.check_output(command, text=True)
        except dockerblade.exceptions.CalledProcessError as error:
            if error.returncode == 2:
                raise exceptions.ROSWireException('illegal service call args') from error  # noqa
            raise exceptions.ROSWireException('unexpected error during service call') from error  # noqa

        fmt_response: Optional[MsgFormat] = self.format.response
        if not fmt_response:
            return None

        d = yaml.safe_load(output)
        db_type = self._description.types
        return db_type.from_dict(fmt_response, d)


class ServiceManager(Mapping[str, Service]):
    """Provides access to the registered services on a ROS graph."""
    def __init__(self,
                 description: SystemDescription,
                 host_ip_master: str,
                 api: xmlrpc.client.Server,
                 shell: dockerblade.Shell
                 ) -> None:
        self.__description = description
        self.__host_ip_master = host_ip_master
        self.__api = api
        self.__shell = shell
        self.__state_probe: SystemStateProbe = \
            SystemStateProbe.via_xmlrpc_connection(self.__api)

    def __get_service_names(self) -> AbstractSet[str]:
        return set(self.__state_probe().services.keys())

    def __len__(self) -> int:
        """The number of advertised services on this ROS graph."""
        return len(self.__get_service_names())

    def __iter__(self) -> Iterator[str]:
        """Returns an iterator over the names of all registered services."""
        yield from self.__get_service_names()

    def __getitem__(self, name: str) -> Service:
        """Fetches a proxy for a service with a given name.

        Parameters
        ----------
        name: str
            The name of the service.

        Returns
        -------
        Service
            A proxy to the given service.

        Raises
        ------
        ServiceNotFound
            If no service is found with the given name.
        """
        code: int
        msg: str
        url_container: str
        code, msg, url_container = \
            self.__api.lookupService('/.roswire', name)  # type: ignore

        if code == -1:
            raise exceptions.ServiceNotFoundError(name)
        if code != 1:
            m = "an unexpected error occurred when retrieving services"
            m = f"{m}: {msg} (code: {code})"
            raise exceptions.ROSWireException(m)

        # convert URL to host network
        parsed = urlparse(url_container)
        url_host = f"{parsed.scheme}://{self.__host_ip_master}:{parsed.port}"

        # find the format for the service
        command = f'rosservice type {name}'
        try:
            name_fmt = self.__shell.check_output(command, text=True)
        except dockerblade.exceptions.CalledProcessError as error:
            m = f"unable to determine type for service [{name}]"
            raise exceptions.ROSWireException(m) from error
        fmt = self.__description.formats.services[name_fmt]
        return Service(name=name,
                       url=url_host,
                       format=fmt,
                       description=self.__description,
                       shell=self.__shell)



# -*- coding: utf-8 -*-
__all__ = ('tuple_from_iterable', 'Stopwatch',)

from typing import Tuple, Iterable, Any
from timeit import default_timer as timer
import warnings


def tuple_from_iterable(val: Iterable[Any]) -> Tuple[Any, ...]:
    """Builds a tuple from an iterable.
    Workaround for https://github.com/python-attrs/attrs/issues/519
    """
    return tuple(val)


class Stopwatch:
    def __init__(self) -> None:
        """Constructs a new, paused timer."""
        self.__offset: float = 0.0
        self.__paused: bool = True
        self.__time_start: float = 0.0

    def stop(self) -> None:
        """Freezes the timer."""
        if not self.__paused:
            self.__offset += timer() - self.__time_start
            self.__paused = True

    def start(self) -> None:
        """Resumes the timer."""
        if self.__paused:
            self.__time_start = timer()
            self.__paused = False
        else:
            warnings.warn("timer is already running")

    def reset(self) -> None:
        """Resets and freezes the timer."""
        self.__offset = 0.0
        self.__paused = True

    @property
    def paused(self) -> bool:
        """Returns True if this stopwatch is paused, or False if not."""
        return self.__paused

    @property
    def duration(self) -> float:
        """The number of seconds that the stopwatch has been running."""
        d = self.__offset
        if not self.__paused:
            d += timer() - self.__time_start
        return d



import os
from glob import glob
from setuptools import setup, find_packages


path = os.path.join(os.path.dirname(__file__), 'src/roswire/version.py')
with open(path, 'r') as f:
    exec(f.read())


setup(version=__version__)



# -*- coding: utf-8 -*-
__all__ = ('TCPROSHeader', 'TCPROSMessage')

from io import BytesIO
from typing import Dict, Optional, BinaryIO, Mapping, Type

import attr

from ..definitions.msg import Message
from ..definitions.encode import write_uint32, write_encoded_header
from ..definitions.decode import read_uint32, read_string_dictionary

_UTF8_ONE = '1'.encode('utf-8')


@attr.s(frozen=True, auto_attribs=True, slots=True)
class TCPROSHeader:
    """Stores information for a TCPROS header.

    Attributes
    ----------
    callerid: str
        The name of the node that is sending data.
    md5sum: str
        The md5sum of the associated message type.
    type_: str
        The name of the associated message type.
    message_definition: str
        The full text of the associated message definition.
    error: str, optional
        An optional human-readable error message if the connection is not
        successful
    persistent: bool, optional
        Sent from a service client to a service. If :code:`True`, the
        connection will be kept open for multiple requests.
    tcp_nodelay: bool, optional
        Sent from a subscriber to a publisher. If :code:`True`, the publisher
        will set TCP_NODELAY on the socket if possible.
    latching: bool, optional
        Indicates that the publisher is in latching mode. I.e., the publisher
        sends the last value that it published to any new subscribers.
    topic: str, optional
        The name of the topic, if any, to which the subscriber is connected.
    service: str, optional
        The name of service, if any, that is being called.
    """
    callerid: str
    md5sum: str
    type_: str
    message_definition: str
    tcp_nodelay: Optional[bool] = attr.ib(default=None)
    persistent: Optional[bool] = attr.ib(default=None)
    latching: Optional[bool] = attr.ib(default=None)
    error: Optional[str] = attr.ib(default=None)
    topic: Optional[str] = attr.ib(default=None)
    service: Optional[str] = attr.ib(default=None)

    def write(self, b: BinaryIO) -> None:
        """Writes a binary encoding of this header to a given stream."""
        field_to_bytes: Dict[str, bytes] = {
            'callerid': self.callerid.encode('utf-8'),
            'md5sum': self.md5sum.encode('utf-8')}
        if self.message_definition:
            field_to_bytes['message_definition'] = \
                self.message_definition.encode('utf-8')
        if self.type_:
            field_to_bytes['type'] = self.type_.encode('utf-8')
        if self.service:
            field_to_bytes['service'] = self.service.encode('utf-8')
        if self.topic:
            field_to_bytes['topic'] = self.topic.encode('utf-8')
        if self.error:
            field_to_bytes['error'] = self.error.encode('utf-8')
        if self.tcp_nodelay:
            field_to_bytes['tcp_nodelay'] = _UTF8_ONE
        if self.persistent:
            field_to_bytes['persistent'] = _UTF8_ONE
        if self.latching:
            field_to_bytes['latching'] = _UTF8_ONE
        write_encoded_header(field_to_bytes, b)

    def encode(self) -> bytes:
        """Returns a binary encoding of this header."""
        b = BytesIO()
        self.write(b)
        return b.getvalue()

    @classmethod
    def decode(cls, b: bytes) -> 'TCPROSHeader':
        return cls.read(BytesIO(b))

    @classmethod
    def read(cls, b: BinaryIO) -> 'TCPROSHeader':
        field_to_value: Dict[str, str] = read_string_dictionary(b)

        def fetch_optional_bool(key: str) -> Optional[bool]:
            if key not in field_to_value:
                return None
            return field_to_value[key] == '1'

        callerid = field_to_value['callerid']
        md5sum = field_to_value['md5sum']
        type_ = field_to_value['type']
        message_definition = field_to_value['message_definition']
        service = field_to_value.get('service')
        topic = field_to_value.get('topic')
        error = field_to_value.get('error')
        tcp_nodelay = fetch_optional_bool('tcp_nodelay')
        persistent = fetch_optional_bool('persistent')
        latching = fetch_optional_bool('latching')

        return TCPROSHeader(callerid=callerid,
                            md5sum=md5sum,
                            type_=type_,
                            message_definition=message_definition,
                            tcp_nodelay=tcp_nodelay,
                            persistent=persistent,
                            latching=latching,
                            error=error,
                            topic=topic,
                            service=service)


@attr.s(auto_attribs=True, slots=True)
class TCPROSMessage:
    header: TCPROSHeader
    message: Message

    def encode(self) -> bytes:
        """Returns a binary encoding of this message."""
        b = BytesIO()
        self.write(b)
        return b.getvalue()

    def write(self, b: BytesIO) -> None:
        """Writes this message to a binary stream."""
        self.header.write(b)

        # write a placeholder size
        position_message_size = b.tell()
        write_uint32(0, b)
        position_message_data = b.tell()
        self.message.write(b)
        position_message_end = b.tell()

        # compute the true size and update
        size_message = position_message_end - position_message_data
        b.seek(position_message_size)
        write_uint32(size_message, b)
        b.seek(position_message_end)

    @classmethod
    def decode(cls,
               types: Mapping[str, Type[Message]],
               b: bytes
               ) -> 'TCPROSMessage':
        """Decodes a TCPROS message from its binary form.

        Parameters
        ----------
        types: Mapping[str, Type[Message]]
            A collection of types indexed by name.
        b: bytes
            A binary blob.
        """
        return cls.read(types, BytesIO(b))

    @classmethod
    def read(cls,
             types: Mapping[str, Type[Message]],
             b: BytesIO
             ) -> 'TCPROSMessage':
        """Reads a TCPROS message from a binary stream."""
        header = TCPROSHeader.read(b)
        type_ = types[header.type_]
        # read message size
        read_uint32(b)
        message = type_.read(b)
        return TCPROSMessage(header=header,
                             message=message)



# -*- coding: utf-8 -*-
import pytest

import time


@pytest.mark.parametrize('sut', ['fetch'], indirect=True)
def test_read(sut):
    with sut.roscore() as ros:
        config = ros.roslaunch.read('pickplace_playground.launch', package='fetch_gazebo')
        assert '/ros_ws/src/fetch_gazebo/fetch_gazebo/models:' in config.envs['GAZEBO_MODEL_PATH'].value

        assert config.params['/arm_controller/follow_joint_trajectory/type'].value == 'robot_controllers/FollowJointTrajectoryController'
        assert config.params['/gazebo/bellows_joint/position/i_clamp'].value == 0.0

        actual_node_names = {node.name for node in config.nodes}
        expected_node_names = {'prepare_robot',
                               'gazebo',
                               'crop_decimate',
                               'urdf_spawner',
                               'head_camera_nodelet_manager',
                               'gazebo_gui',
                               'robot_state_publisher',
                               'points_downsample',
                               'cmd_vel_mux'}
        assert actual_node_names == expected_node_names


@pytest.mark.parametrize('sut', ['husky'], indirect=True)
def test_eval_args_in_launch_file(sut):
    with sut.roscore() as ros:
        config = ros.roslaunch.read('spawn_husky.launch', package='husky_gazebo')
        actual_node_names = {node.name for node in config.nodes}
        expected_node_names = {'base_controller_spawner',
                               'ekf_localization',
                               'robot_state_publisher',
                               'spawn_husky_model',
                               'twist_marker_server',
                               'twist_mux'}
        assert actual_node_names == expected_node_names


@pytest.mark.parametrize('sut', ['fetch'], indirect=True)
def test_remappings(sut):
    with sut.roscore() as ros:
        launch = ros.roslaunch('pickplace_playground.launch',
                               package='fetch_gazebo',
                               node_to_remappings={'gazebo': [('/gazebo/model_states', '/funkybits')]})

        time.sleep(30)
        state = ros.state
        expected_nodes = {'/cmd_vel_mux',
                          '/gazebo',
                          '/head_camera/crop_decimate',
                          '/head_camera/depth_downsample/points_downsample',
                          '/head_camera/head_camera_nodelet_manager',
                          '/robot_state_publisher',
                          '/rosout'}
        assert state.nodes == expected_nodes
        published_topics = set(state.publishers)
        assert '/gazebo/model_states' not in state.publishers
        assert set(state.publishers['/funkybits']) == {'/gazebo'}



import pytest

import roswire.name


def test_global_name():
    f = roswire.name.global_name
    assert f('') == '/'
    assert f('/') == '/'
    assert f('foo') == '/foo/'
    assert f('foo/') == '/foo/'
    assert f('/foo/') == '/foo/'
    assert f('foo/bar') == '/foo/bar/'
    assert f('foo/bar/') == '/foo/bar/'
    assert f('/foo/bar/') == '/foo/bar/'


def test_canonical_name():
    f = roswire.name.canonical_name
    assert f('') == ''
    assert f('/') == '/'
    assert f('foo') == 'foo'
    assert f('foo/') == 'foo'
    assert f('/foo') == '/foo'
    assert f('/foo/') == '/foo'
    assert f('foo/bar') == 'foo/bar'
    assert f('foo/bar/') == 'foo/bar'
    assert f('/foo/bar/') == '/foo/bar'


def test_name_is_private():
    f = roswire.name.name_is_private
    assert not f('')
    assert not f('/')
    assert not f('/foo')
    assert not f('/foo/bar')
    assert not f('foo')
    assert not f('foo/')
    assert f('~foo')
    assert f('~bar')


def test_name_is_global():
    f = roswire.name.name_is_global
    assert f('/')
    assert f('/foo')
    assert f('/foo/bar')
    assert not f('')
    assert not f('foo')
    assert not f('foo/')
    assert not f('~foo')
    assert not f('~bar')


def test_name_is_legal():
    f = roswire.name.name_is_legal
    assert f('')
    assert f('/')
    assert f('/foo')
    assert f('/foo/bar')
    assert f('/foo/bar/')
    assert f('~')
    assert f('~foo')
    assert not f('/~foo')
    assert not f('foo bar')


def test_namespace():
    f = roswire.name.namespace
    assert f('/') == '/'
    assert f('') == '/'
    assert f('/bar') == '/'
    assert f('/bar/foo') == '/bar'
    with pytest.raises(ValueError):
        f('~foo')


def test_namespace_join():
    f = roswire.name.namespace_join
    assert f('', 'foo') == 'foo'
    assert f('', 'foo/bar') == 'foo/bar'
    assert f('/', 'foo') == '/foo'
    assert f('/foo', 'bar') == '/foo/bar'
    assert f('/foo/bar', 'beep') == '/foo/bar/beep'
    assert f('/foo/bar', '~beep') == '~beep'
    assert f('/foo/bar', '/bork') == '/bork'



# -*- coding: utf-8 -*-
"""
This module provides data structures for representing roslaunch configurations.
"""
from .launch import LaunchConfig
from .node import NodeConfig
from .parameter import Parameter



# -*- coding: utf-8 -*-
__all__ = ('Parameter',)

from typing import Any, Optional
import xml.etree.ElementTree as ET

import attr
import yaml


@attr.s(frozen=True, slots=True, auto_attribs=True)
class Parameter:
    """Provides a ROS parameter definition.

    Attributes
    ----------
    name: str
        The name of the parameter.
    typ: str
        The name of the type used by the parameter.
    command: str, optional
        The command that was used to define the parameter, if known.
    value: Any
        The value of the parameter.
    """
    name: str
    typ: str
    value: Any
    command: Optional[str] = attr.ib(default=None)

    def to_xml_element(self) -> ET.Element:
        # primitive parameters
        if type(self.value) in (str, float, int, bool):
            element = ET.Element('param')
            element.attrib['name'] = self.name

            if self.typ != 'auto':
                element.attrib['type'] = self.typ

            if self.command:
                element.attrib['command'] = self.command
            elif isinstance(self.value, bool):
                element.attrib['value'] = 'true' if self.value else 'false'
            else:
                element.attrib['value'] = str(self.value)
            return element

        # complex parameters
        element = ET.Element('rosparam')
        element.text = yaml.safe_dump(self.value, default_flow_style=True)
        element.attrib['param'] = self.name
        return element



# -*- coding: utf-8 -*-
__all__ = ('Container', 'ContainerManager')

from typing import Dict, Iterator, Mapping, Optional, Sequence, Union
from uuid import UUID, uuid4
import contextlib
import shutil
import os

import attr
import dockerblade
from docker import DockerClient
from docker import APIClient as DockerAPIClient
from docker.models.images import Image as DockerImage
from docker.models.containers import Container as DockerContainer
from loguru import logger

from ..exceptions import SourceNotFoundError


@attr.s(frozen=True)
class Container:
    """Provides an interface to a Docker container running on this host.

    Attributes
    ----------
    dockerblade: dockerblade.container.Container
        Provides access to the underlying Docker container via Dockerblade.
    sources: Sequence[str]
        The sequence of setup files that should be used to load the ROS
        workspace.
    environment: Mapping[str, str]
        A set of additional environment variables, indexed by name, that should
        be used by the container.
    uuid: UUID
        A unique identifier for this container.
    shell: dockerblade.shell.Shell
        Provides access to a bash shell for this container.
    files: dockerblade.files.FileSystem
        Provides access to the filesystem for this container.
    pid: int
        The PID of this container process on the host machine.
    ws_host: str
        The absolute path to the shared directory for this container's
        workspace on the host machine.
    ip_address: str
        The IP address for this container on the host network.

    Raises
    ------
    SourceNotFoundError:
        If at least one of the given sources could not be found inside the
        container.
    """
    _dockerblade: dockerblade.container.Container = attr.ib(repr=False)
    _sources: Sequence[str] = attr.ib(repr=False)
    uuid: UUID = attr.ib(repr=True)
    ws_host: str = attr.ib(repr=False)
    _environment: Mapping[str, str] = attr.ib(repr=False, factory=dict)
    shell: dockerblade.shell.Shell = attr.ib(init=False, repr=False)
    files: dockerblade.files.FileSystem = attr.ib(init=False, repr=False)

    def __attrs_post_init__(self) -> None:
        files = self._dockerblade.filesystem()

        # #338 ensure that the given sources exist
        for source in self._sources:
            if not files.exists(source):
                raise SourceNotFoundError(source)

        shell = self._dockerblade.shell('/bin/bash',
                                        sources=self._sources,
                                        environment=self._environment)
        object.__setattr__(self, 'shell', shell)
        object.__setattr__(self, 'files', files)

    @property
    def pid(self) -> int:
        return self._dockerblade.pid

    @property
    def ip_address(self) -> str:
        ip_address = self._dockerblade.ip_address
        assert ip_address
        return ip_address

    def persist(self,
                repo: Optional[str] = None,
                tag: Optional[str] = None
                ) -> DockerImage:
        """Persists this container to an image.

        Parameters
        ----------
        repo: str, optional
            The name of the repository to which the image should belong.
        tag: str, optional
            The tag that should be used for the image.

        Returns
        -------
        DockerImage
            A description of the persisted image.
        """
        return self._dockerblade.persist(repo, tag)


@attr.s(auto_attribs=True)
class ContainerManager:
    """
    Provides an interface for accessing and inspecting Docker images, and
    launching Docker containers.
    """
    _dir_host_workspace: str
    _dockerblade: dockerblade.daemon.DockerDaemon = \
        attr.ib(factory=dockerblade.daemon.DockerDaemon)

    @property
    def docker_client(self) -> DockerClient:
        return self._dockerblade.client

    @property
    def docker_api(self) -> DockerAPIClient:
        return self._dockerblade.api

    @contextlib.contextmanager
    def _build_shared_directory(self, uuid: UUID) -> Iterator[str]:
        dir_host = os.path.join(self._dir_host_workspace,
                                'containers',
                                uuid.hex)
        try:
            logger.debug(f"creating container directory: {dir_host}")
            os.makedirs(dir_host, exist_ok=True)
            logger.debug(f"created container directory: {dir_host}")
            yield dir_host
        finally:
            if os.path.exists(dir_host):
                logger.debug(f"destroying container directory: {dir_host}")
                shutil.rmtree(dir_host)
                logger.debug(f"destroyed container directory: {dir_host}")

    @contextlib.contextmanager
    def _container(self,
                   image_or_name: Union[str, DockerImage],
                   dir_host_shared: str,
                   uuid: UUID,
                   *,
                   ports: Optional[Dict[int, int]] = None
                   ) -> Iterator[DockerContainer]:
        logger.debug(f"building docker container: {uuid}")
        dockerc = self.docker_client.containers.create(
            image_or_name,
            '/bin/sh',
            user='root',
            name=uuid,
            entrypoint='/bin/sh -c',
            volumes={dir_host_shared: {'bind': '/.roswire', 'mode': 'rw'}},
            ports=ports,
            stdin_open=True,
            tty=False,
            detach=True)
        logger.debug("created docker container")
        try:
            dockerc.start()
            logger.debug(f"started docker container: {uuid}")
            yield dockerc
        finally:
            logger.debug(f"destroying docker container: {uuid}")
            dockerc.remove(force=True)
            logger.debug(f"destroyed docker container: {uuid}")

    @contextlib.contextmanager
    def launch(self,
               image_or_name: Union[str, DockerImage],
               sources: Sequence[str],
               *,
               ports: Optional[Dict[int, int]] = None,
               environment: Optional[Mapping[str, str]] = None
               ) -> Iterator['Container']:
        """
        Launches a context-managed Docker container for a given image. Upon
        exiting the context, whether gracefully or abruptly (i.e., via an
        unhandled exception), the container will be automatically destroyed.

        Parameters
        ----------
        image_or_name: Union[str, DockerImage]
            The image that should be used to create the container, or the name
            of that image.
        sources: Sequence[str]
            The sequence of setup files that should be used to load the ROS
            workspace.
        ports: Dict[int, int], optional
            An optional dictionary specifying port mappings between the host
            and container, where keys represent container ports and values
            represent host ports.
        environment: Mapping[str, str], optional
            An optional set of additional environment variables, indexed by
            name, that should be used by the container.

        Yields
        ------
        Container
            The constructed container.
        """
        uuid = uuid4()
        logger.debug(f"UUID for container: {uuid}")
        if not environment:
            environment = {}
        with contextlib.ExitStack() as stack:
            dir_shared = stack.enter_context(self._build_shared_directory(uuid))  # noqa
            dockerc = stack.enter_context(self._container(image_or_name, dir_shared, uuid, ports=ports))  # noqa
            dockerb = self._dockerblade.attach(dockerc.id)
            yield Container(dockerblade=dockerb,
                            sources=sources,
                            uuid=uuid,
                            ws_host=dir_shared,
                            environment=environment)

    def image(self, tag: str) -> DockerImage:
        """Retrieves the Docker image with a given tag."""
        return self.docker_client.images.get(tag)

    def image_sha256(self, tag_or_image: Union[str, DockerImage]) -> str:
        """Computes the SHA256 for a given Docker image."""
        image: DockerImage
        if isinstance(tag_or_image, str):
            image = self.image(tag_or_image)
        else:
            image = tag_or_image
        return image.id[7:]



# -*- coding: utf-8 -*-
"""
This module provides functionality for converting the contents of ROS bag
files into Daikon trace (and declaration) files.

Right now, this module does not handle arrays, timestamps, and durations.

References
----------
https://plse.cs.washington.edu/daikon/download/doc/developer/File-formats.html#Data-trace-records
"""
__all__ = ('bag_to_decls', 'bag_to_daikon')

from typing import Set, FrozenSet, List, Iterator, Collection, Mapping, Any
from collections import OrderedDict
from functools import reduce

import attr

from .reader import BagReader
from ..description import SystemDescription
from ..definitions import Message, MsgFormat


TYPE_TO_DAIKON = {
    'bool': 'boolean',
    'int8': 'byte',
    'uint8': 'byte',
    'int16': 'short',
    'uint16': 'short',
    'int32': 'int',
    'uint32': 'int',
    'int64': 'long',
    'uint64': 'long',
    'float32': 'double',
    'float64': 'double',
    'string': 'java.lang.String'}


@attr.s(frozen=True, str=False)
class VarDecl:
    name: str = attr.ib()
    dec_type: str = attr.ib()
    rep_type: str = attr.ib()

    @property
    def lines(self) -> List[str]:
        return [f'variable {self.name}',
                '  var-kind variable',
                f'  dec-type {self.dec_type}',
                f'  rep-type {self.rep_type}']

    def __str__(self) -> str:
        return '\n'.join(self.lines)


@attr.s(frozen=True, str=False)
class GenericProgramPoint:
    name: str = attr.ib()
    variables: FrozenSet[VarDecl] = attr.ib(converter=frozenset)

    @property
    def lines(self) -> List[str]:
        ls = [f'ppt {self.fullname}', 'ppt-type point']
        for var in self.variables:
            ls += var.lines
        return ls

    @property
    def fullname(self) -> str:
        return f'{self.name}:::POINT'

    def __str__(self) -> str:
        return '\n'.join(self.lines)


class Declarations(Mapping[str, GenericProgramPoint]):
    def __init__(self, points: Collection[GenericProgramPoint]) -> None:
        self.__points = OrderedDict((p.name, p) for p in points)

    def __len__(self) -> int:
        return len(self.__points)

    def __getitem__(self, name: str) -> GenericProgramPoint:
        return self.__points[name]

    def __iter__(self) -> Iterator[str]:
        yield from self.__points

    @property
    def lines(self) -> List[str]:
        ls = ['decl-version 2.0', 'var-comparability none']
        for ppt in self.values():
            ls += ppt.lines
        return ls

    def __str__(self) -> str:
        return '\n'.join(self.lines)

    def save(self, filename: str) -> None:
        with open(filename, 'w') as f:
            f.write(str(self))


class TraceWriter:
    """Used to write to Daikon trace files."""
    def __init__(self, filename: str) -> None:
        self.__filename = filename
        self.__fp = open(filename, 'w')

    def add(self, ppt: GenericProgramPoint, vals: Mapping[str, Any]) -> None:
        """Writes an entry to the trace file."""
        fp = self.__fp
        lines: List[str] = [ppt.fullname]
        for var_name, var_val in vals.items():
            if isinstance(var_val, bool):
                str_val = 'true' if var_val else 'false'
            elif isinstance(var_val, str):
                str_val = var_val.replace('\\', '\\\\')
                str_val = str_val.replace('"', '\\"')
                str_val = f'"{str_val}"'
            else:
                str_val = str(var_val)
            lines += [var_name, str_val, '1']
        fp.writelines(f'{line}\n' for line in lines)
        fp.write('\n')

    def close(self) -> None:
        self.__fp.close()


def topic_to_ppt(topic_name: str,
                 topic_fmt: MsgFormat,
                 sys_desc: SystemDescription
                 ) -> GenericProgramPoint:
    """Creates a program point for a given ROS topic."""
    decls: Set[VarDecl] = set()
    for field_ctx, field in topic_fmt.flatten(sys_desc.formats.messages):
        if field.typ not in TYPE_TO_DAIKON:
            continue

        field_name = field.name
        if field_ctx:
            field_name = f"{'.'.join(field_ctx)}.{field_name}"

        field_type = TYPE_TO_DAIKON[field.typ]
        dec_type = field_type
        rep_type = field_type
        decl = VarDecl(field_name, dec_type, rep_type)
        decls.add(decl)
    return GenericProgramPoint(topic_name, decls)  # type: ignore


def bag_to_decls(fn_bag: str, sys_desc: SystemDescription) -> Declarations:
    """Builds a .decls file for a given ROS bag.

    Parameters
    ----------
    fn_bag: str
        the path to the bag file.
    sys_desc: SystemDescription
        a description of the system used to produce the bag.

    Returns
    -------
    Declarations:
        a declarations description.
    """
    # determine the set of topics (and their types) represented in the bag
    reader = BagReader(fn_bag, sys_desc.types)
    topic_to_type = reader.topics_to_types

    # transform each topic to a program point
    ppts: Set[GenericProgramPoint] = set()
    for topic_name, topic_type in topic_to_type.items():
        ppt = topic_to_ppt(topic_name, topic_type.format, sys_desc)
        ppts.add(ppt)

    return Declarations(ppts)  # type: ignore


def bag_to_daikon(fn_bag: str,
                  fn_dtrace: str,
                  fn_decls: str,
                  sys_desc: SystemDescription
                  ) -> None:
    """Transforms a ROS bag into a Daikon trace.

    Parameters
    ----------
    fn_bag: str
        the path to the bag file.
    fn_dtrace: str
        the path to the output Daikon trace file.
    fn_decls: str
        the path to the output Daikon decls file.
    sys_desc: SystemDescription
        a description of the system used to produce the bag.
    """
    decls = bag_to_decls(fn_bag, sys_desc)
    decls.save(fn_decls)

    def read_val(m: Message, name_field: str) -> Any:
        return reduce(getattr, name_field.split('.'), m)

    trace_writer = TraceWriter(fn_dtrace)
    bag_reader = BagReader(fn_bag, sys_desc.types)
    for message in bag_reader:
        ppt = decls[message.topic]
        vals = {}
        for var_decl in ppt.variables:
            vals[var_decl.name] = read_val(message.message, var_decl.name)
        trace_writer.add(ppt, vals)
